@@ -12,7 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-SCHEMA_VERSION = 70  # remember to update the list below when updating
+SCHEMA_VERSION = 71  # remember to update the list below when updating
 """Represents the expectations made by the codebase about the database schema
 
 This should be incremented whenever the codebase changes its requirements on the
@@ -67,6 +67,10 @@ Changes in SCHEMA_VERSION = 69:
 
 Changes in SCHEMA_VERSION = 70:
     - event_reference_hashes is no longer written to.
+
+Changes in SCHEMA_VERSION = 71:
+    - event_edges.room_id is no longer read from.
+    - Tables related to groups are no longer accessed.
 """
 
 
